"""Functions to merge datasets."""

from collections.abc import Iterable
from typing import Protocol, cast

import numpy as np

from allisbns.dataset import CodeDataset, UnpackedCodes


class MergeInPlaceFunction(Protocol):
    """Typing protocol for merge functions with in-place evaluation support.

    The protocol is compatible with NumPy's logical operation `functions
    <https://numpy.org/devdocs/reference/routines.logic.html#logical-operations>`__
    that take the ``out`` argument.

    """

    def __call__(
        self, a: UnpackedCodes, b: UnpackedCodes, /, out: UnpackedCodes | None
    ) -> UnpackedCodes: ...


class MergePureFunction(Protocol):
    """Typing protocol for merge out-of-place functions."""

    def __call__(self, a: UnpackedCodes, b: UnpackedCodes) -> UnpackedCodes: ...


def merge(
    datasets: Iterable[CodeDataset],
    function: MergeInPlaceFunction | MergePureFunction,
) -> CodeDataset:
    """Merges datasets using a merge function.

    The boundaries of all datasets should be the same.

    Arguments:
        datasets: Datasets to merge.
        function: A merge function.

    Returns:
        A merged dataset.

    See Also:
        :func:`union`, :func:`intersection`, :func:`difference`,
        :func:`symmetric_difference`
    """
    dataset_iterator = iter(datasets)

    initial = next(dataset_iterator)
    merged_unpacked = initial.unpack_codes()

    for other in dataset_iterator:
        if other.bounds != initial.bounds:
            raise ValueError("dataset boundaries must be the same")
        try:
            cast("MergeInPlaceFunction", function)(
                merged_unpacked, other.unpack_codes(), out=merged_unpacked
            )
        except TypeError:
            merged_unpacked[:] = cast("MergePureFunction", function)(
                merged_unpacked, other.unpack_codes()
            )

    return CodeDataset.from_unpacked(merged_unpacked, offset=initial.offset)


def union(x: Iterable[CodeDataset]) -> CodeDataset:
    return merge(x, np.logical_or)


def intersection(x: Iterable[CodeDataset]) -> CodeDataset:
    return merge(x, np.logical_and)


def difference(x: Iterable[CodeDataset]) -> CodeDataset:
    def merge_function(
        a: UnpackedCodes, b: UnpackedCodes, /, out: UnpackedCodes | None
    ) -> UnpackedCodes:
        return np.logical_and(a, np.invert(b, out=b), out=out)

    return merge(x, merge_function)


def symmetric_difference(x: CodeDataset, y: CodeDataset) -> CodeDataset:
    return merge([x, y], np.logical_xor)



"""Exceptions specific to the package."""

from __future__ import annotations

from enum import StrEnum, auto
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    import sys

    if "sphinx.ext.autodoc" in sys.modules:
        from allisbns.isbn import CanonicalISBN, MaybeISBN


class AllISBNsError(Exception):
    """Base package exception."""


class InvalidISBNReason(StrEnum):
    """Reasons for invalid ISBNs.

    Supposed to be used for :data:`allisbns.isbn.MaybeISBN` and
    :class:`allisbns.isbn.CanonicalISBN`.

    See Also:
        :data:`allisbns.isbn.ISBN12`

    Examples:
        While '978-0-00-000000-K' contains the bad check digit ('K'), we first
        don't recognize it as an ISBN with :func:`~allisbns.isbn.match_isbn`,
        and therefore :attr:`~InvalidISBNReason.NOT_ISBN` should be used
        here. Such reasons (along with bad length, bad characters, etc.) could
        be taken out into a separate enum, say, ``NotISBNReason``, but we don't
        go that far.
    """

    #: Not an ISBN.
    NOT_ISBN = auto()
    #: Undefined registration group.
    BAD_GROUP = auto()
    #: Undefined registrant.
    BAD_REGISTRANT = auto()
    #: Incorrectly calculated check digit.
    BAD_CHECK_DIGIT = auto()


class InvalidISBNError(AllISBNsError, ValueError):
    """Raised when an ISBN is not valid."""

    def __init__(
        self, message: str, reason: InvalidISBNReason, isbn: MaybeISBN | CanonicalISBN
    ):
        """Creates an exception with ``message`` and ``reason`` for ``isbn``."""
        super().__init__(message)
        self.message = message
        self.reason = reason
        self.isbn = isbn

        self.add_note(f"\nProblematic ISBN: '{self.isbn}'")

    def __str__(self) -> str:
        return f"{self.reason}: {self.message}"


class NotISBNError(InvalidISBNError):
    """Raised when the value is not an ISBN."""

    def __init__(self, isbn: MaybeISBN, message: str | None = None):
        """Creates an exception with ``message`` for ``isbn``."""
        message = message or "value is neither ISBN-10 nor ISBN-13"
        super().__init__(message, InvalidISBNReason.NOT_ISBN, isbn)

    def __str__(self) -> str:
        return self.message


class UndefinedISBNRangeError(InvalidISBNError):
    """Raised when the registration group or registrant range are undefined."""

    def __init__(self, message: str, reason: InvalidISBNReason, isbn: CanonicalISBN):
        """Creates an exception with ``message`` and ``reason`` for ``isbn``."""
        super().__init__(message, reason, isbn)

        self.add_note(
            "\nNote: might be caused due to the ranges file expiration. "
            "Make sure that the range is defined for use"
        )


class NotISBN12Error(AllISBNsError, ValueError):
    """Raised when a value is not an ISBN-12 number."""



"""Supports plotting of binned images."""

import math

from abc import ABC, abstractmethod
from typing import Any, Literal, cast

import matplotlib as mpl
import matplotlib.typing as mpt
import numpy as np
import numpy.typing as npt

from matplotlib.axes import Axes

from allisbns.dataset import BinnedArray
from allisbns.isbn import FIRST_ISBN, ISBN12, LAST_ISBN
from allisbns.rearrange import rearrange_to_blocks, rearrange_to_rows


type MatplotlibExtent = tuple[float, float, float, float]


def truncate_colormap(
    colormap, vmin: float = 0, vmax: float = 0.98
) -> mpl.colors.Colormap:
    return mpl.colors.ListedColormap(
        colormap(np.linspace(vmin, vmax, colormap.N)), f"{colormap.name}_truncated"
    )


def get_default_colormap(
    nan_color: mpt.ColorType = "0.6", under_color: mpt.ColorType = "0.45"
) -> mpl.colors.Colormap:
    colormap = truncate_colormap(mpl.colormaps.get_cmap("plasma"))
    colormap.set_bad(nan_color)
    colormap.set_under(under_color)
    return colormap


def tweak_colormap(
    colormap: mpl.colors.Colormap,
    nan_color: mpt.ColorType | None = None,
    under_color: mpt.ColorType | None = None,
) -> mpl.colors.Colormap:
    if nan_color:
        colormap.set_bad(nan_color)
    if under_color:
        colormap.set_under(under_color)
    return colormap


def translate_with_padding(ax, transform, *, x: float = 0, y: float = 0):
    physical_transform = mpl.transforms.ScaledTranslation(
        x, y, ax.get_figure().dpi_scale_trans
    )
    return transform + physical_transform


def pad_along_x(ax, value: float):
    transform = mpl.transforms.blended_transform_factory(ax.transAxes, ax.transData)
    return translate_with_padding(ax, transform, x=value)


def pad_along_y(ax, value: float):
    transform = mpl.transforms.blended_transform_factory(ax.transData, ax.transAxes)
    return translate_with_padding(ax, transform, y=value)


class CoordinateConverter(ABC):
    """Represents a coordinate converter for bins rearranged in various ways."""

    @abstractmethod
    def xy_to_isbn(self, x: int, y: int) -> ISBN12:
        """Converts coordinates to an ISBN."""

    @abstractmethod
    def isbn_to_xy(self, isbn: ISBN12) -> tuple[int, int]:
        """Converts an ISBN to coordinates."""


class RowCoordinateConverter(CoordinateConverter):
    """Represents a coordinate converter for bins rearranged into rows."""

    def __init__(self, width: int, bin_size: int, offset: ISBN12):
        """Creates a coordinate converter.

        Arguments:
            width: A width of rows.
            bin_size: A size of bins.
            offset: An ISBN offset.
        """
        self.width = int(width)
        self.bin_size = int(bin_size)
        self.offset = int(offset)

    def xy_to_isbn(self, x: int, y: int) -> ISBN12:
        """Converts image coordinates to an ISBN."""
        return self.offset + y * self.width + x * self.bin_size

    def isbn_to_xy(self, isbn: ISBN12) -> tuple[int, int]:
        """Converts an ISBN to image coordinates."""
        bin_index = (isbn - self.offset) // self.bin_size
        y, x = divmod(bin_index, self.width // self.bin_size)
        return x, y


class BlockCoordinateConverter(CoordinateConverter):
    """Represents a coordinate converter for bins rearranged into blocks."""

    def __init__(
        self, block_width: int, block_size: int, bin_size: int, offset: ISBN12
    ):
        """Creates a coordinate converter.

        Arguments:
            block_width: A width of one block.
            block_size: A number of ISBNs in one block.
            bin_size: A size of bins.
            offset: An ISBN offset.
        """
        self.block_width = int(block_width)
        self.block_size = int(block_size)
        self.bin_size = int(bin_size)
        self.offset = offset

        self._bins_per_row = (block_width + bin_size - 1) // bin_size
        self._rows_per_block = (block_size + block_width - 1) // block_width

    def xy_to_isbn(self, x: int, y: int) -> ISBN12:
        block_index, subcolumn_index = divmod(x, self._bins_per_row)

        block_offset = block_index * self.block_size
        row_offset = y * self.block_width
        subcolumn_offset = subcolumn_index * self.bin_size

        return self.offset + block_offset + row_offset + subcolumn_offset

    def isbn_to_xy(self, isbn: ISBN12) -> tuple[int, int]:
        bin_index = (isbn - self.offset) // self.bin_size

        bins_per_block = self._bins_per_row * self._rows_per_block
        block_index, bin_in_block = divmod(bin_index, bins_per_block)
        y, subcolumn_index = divmod(bin_in_block, self._bins_per_row)

        x = block_index * self._bins_per_row + subcolumn_index

        return x, y


class BinnedPlotter(ABC):
    """Represents a base plotter for bins."""

    def __init__(
        self,
        ax,
        *,
        bin_size: int,
        offset: ISBN12 = FIRST_ISBN,
        aspect: Literal["auto", "equal"] | float = "equal",
        decorate_axes: bool = True,
    ):
        """Creates a plotter for bins.

        Arguments:
            ax: An axis into which to plot.
            bin_size: A size of bins.
            offset: An ISBN offset.
            aspect: An aspect of the image.
            decorate_axes: Whether to decorate axes with the plotter's specific
                axis and tick labels.
        """
        self.ax = ax
        self.bin_size = int(bin_size)
        self.offset = offset
        self.aspect = aspect
        self.decorate_axes = decorate_axes

        self._extent: MatplotlibExtent | None = None

        vmin = 20
        self.default_imshow_kwargs = {
            "vmin": min(bin_size, vmin),
            "vmax": bin_size,
            "interpolation": "none",
        }
        self.default_colorbar_kwargs = {
            "ax": self.ax,
            "label": "ISBNs/bin",
            "orientation": "horizontal",
            "location": "top",
            "aspect": 28,
            "shrink": 0.28,
            "anchor": (0, 0),
            "pad": 0.07,
            "extendrect": True,
        }

    @property
    def extent(self) -> MatplotlibExtent | None:
        """The extent of the image as tuple ``(left, right, bottom, top)``."""
        return self._extent

    def _style_axes(self) -> None:
        self.ax.minorticks_on()
        self.ax.yaxis.set_minor_locator(mpl.ticker.AutoMinorLocator(1))

        self.ax.spines[:].set_visible(False)

        self.ax.xaxis.tick_top()
        self.ax.xaxis.set_label_position("top")

        self.ax.xaxis.set_ticks_position("both")
        self.ax.yaxis.set_ticks_position("both")

    @abstractmethod
    def _decorate_axes(self):
        """Decorate axes with the plotter's specific axis and tick labels."""

    def _check_bin_size(self, binned: BinnedArray) -> None:
        if binned.bin_size != self.bin_size:
            raise ValueError(
                f"size of bins ({binned.bin_size}) should match "
                f"the plotter's value ({self.bin_size})"
            )

    def plot_image(
        self,
        image: npt.NDArray,
        *,
        colormap: mpl.colors.Colormap | None = None,
        value_to_nan: int | None = 0,
        show_colorbar: bool = False,
        colorbar_kwargs: dict[str, Any] | None = None,
        **imshow_kwargs,
    ):
        """Plots an image.

        Arguments:
            image: An image array to plot.
            colormap: A colormap.
            value_to_nan: A value that will be replaced by :attr:`numpy:numpy.nan`.
            show_colorbar: Whether to add a colorbar to an image.
            colorbar_kwargs: Additional arguments accepted by
                :func:`~matplotlib:matplotlib.pyplot.colorbar`.
            imshow_kwargs: Additional arguments accepted by
                :func:`~matplotlib:matplotlib.pyplot.imshow`.
        """
        imshow_kwargs = self.default_imshow_kwargs | imshow_kwargs

        if "norm" in imshow_kwargs:
            raise NotImplementedError("Currently, 'norm' argument is not supported")

        vmin = cast("int", imshow_kwargs["vmin"])
        vmax = cast("int", imshow_kwargs["vmax"])
        if vmax is not None and vmin > vmax:
            raise ValueError("vmin cannot be greater than vmax")

        colorbar_kwargs = self.default_colorbar_kwargs | (colorbar_kwargs or {})
        if "extend" not in colorbar_kwargs:
            colorbar_kwargs["extend"] = "min" if imshow_kwargs["vmin"] else "neither"

        image = image.astype(np.float32)

        # Update the extent
        image_height, image_width = image.shape[:2]
        image_extent = (0, image_width, image_height, 0)

        if self.extent is None:
            self._extent = image_extent
        elif self.extent != image_extent:
            raise ValueError(
                f"image extent, {image_extent}, should be equal"
                f"to plotter extent, {self.extent}"
            )

        # Draw the image
        if value_to_nan is not None:
            image[image == value_to_nan] = np.nan

        if colormap is None:
            colormap = get_default_colormap()

        im = self.ax.imshow(
            image,
            cmap=colormap,
            extent=self.extent,
            aspect=self.aspect,
            **imshow_kwargs,
        )

        # Show the colorbar
        if show_colorbar:
            colorbar = self.ax.get_figure().colorbar(im, **colorbar_kwargs)
            vmin = cast("int", imshow_kwargs["vmin"])
            vmax = cast("int", imshow_kwargs["vmax"])

            if vmin > 0:
                colorbar.ax.set_xticks([vmin, *colorbar.ax.get_xticks()[1:]])
                colorbar.ax.set_xticklabels(
                    [f"{vmin:.0f}", *colorbar.ax.get_xticklabels()[1:]]
                )

            colorbar.ax.tick_params(axis="x", direction="in")
            colorbar.ax.set_xlim(vmin, vmax)

        return im

    @abstractmethod
    def plot_bins(
        self,
        binned: BinnedArray,
        colormap: mpl.colors.Colormap | None = None,
        value_to_nan: int | None = 0,
        show_colorbar: bool = True,
        colorbar_kwargs: dict | None = None,
        **imshow_kwargs,
    ):
        pass

    @abstractmethod
    def define_extent(self, end_isbn: ISBN12 = LAST_ISBN) -> None:
        """Defines the extent without plotting bins or an image.

        Arguments:
            end_isbn: An end ISBN to expand an extent to from the plotter offset.
        """


class RowBinnedPlotter(BinnedPlotter):
    """Represents a plotter for bins with a fixed width of rows."""

    def __init__(self, ax: Axes, bin_size: int, width: int = int(2.5e6), **kwargs):
        """Creates a plotter for bins with a fixed width of rows.

        Internally, bins are rearranged into an image with the
        :meth:`~allisbns.rearrange.rearrange_to_rows` method.

        Arguments:
            ax: An axis into which to plot.
            bin_size: A size of bins.
            width: A width of rows.
            kwargs: Additional arguments to pass to :class:`BinnedPlotter`.
        """
        super().__init__(ax, bin_size=bin_size, **kwargs)

        self.width = int(width)
        self.coordinate_converter = RowCoordinateConverter(
            self.width, self.bin_size, self.offset
        )

        self._style_axes()
        if self.decorate_axes:
            self._decorate_axes()

    def _decorate_axes(self):
        xticks_formatter = mpl.ticker.FuncFormatter(
            lambda x, _: f"{x * self.bin_size:,.0f}"
        )
        xticks_formatter.set_offset_string(
            rf"ISBN = {self.offset:,.0f}(X) + $x$ $\times$ ($y$ + 1)"
        )
        self.ax.xaxis.set_major_formatter(xticks_formatter)
        offset_text = self.ax.xaxis.get_offset_text()
        offset_text.set_x(0)
        offset_text.set_horizontalalignment("left")

        self.ax.yaxis.set_major_formatter(
            mpl.ticker.FuncFormatter(lambda x, _: f"{x:.0f}")
        )

        self.ax.xaxis.set_minor_locator(mpl.ticker.AutoMinorLocator(5))

        self.ax.set_xlabel("Relative ISBN, $x$")
        self.ax.set_ylabel("Row, $y$")

    def define_extent(self, end_isbn: ISBN12 = LAST_ISBN) -> None:
        """Defines the extent without plotting bins or an image.

        Arguments:
            end_isbn: An end ISBN to expand an extent to from the plotter offset.
        """
        x, y = self.coordinate_converter.isbn_to_xy(end_isbn)

        self._extent = (0, x + 1, y + 1, 0)
        x1, x2, y1, y2 = self._extent

        self.ax.set_xlim(x1, x2)
        self.ax.set_ylim(y1, y2)

        self.ax.set_aspect(self.aspect)

    def plot_bins(
        self,
        binned: BinnedArray,
        colormap: mpl.colors.Colormap | None = None,
        value_to_nan: int | None = 0,
        show_colorbar: bool = True,
        colorbar_kwargs: dict | None = None,
        **imshow_kwargs,
    ):
        self._check_bin_size(binned)
        image = rearrange_to_rows(binned, binned.bin_size, self.width)
        return self.plot_image(
            image,
            colormap=colormap,
            value_to_nan=value_to_nan,
            show_colorbar=show_colorbar,
            colorbar_kwargs=colorbar_kwargs,
            **imshow_kwargs,
        )


class BlockBinnedPlotter(BinnedPlotter):
    """Represents a plotter for bins stacked as vertical blocks."""

    def __init__(
        self,
        ax: Axes,
        bin_size: int,
        block_width: int = int(1e5),
        block_size: int = int(5e7),
        decorate_axes: bool = True,
        **kwargs,
    ):
        """Creates a plotter for bins stacked as vertical blocks.

        Internally, bins are rearranged into an image with the
        :meth:`~allisbns.rearrange.rearrange_to_blocks` method.

        Arguments:
            ax: An axis into which to plot.
            bin_size: A size of bins.
            block_width: A width of one block.
            block_size: A number of ISBNs in one block.
            decorate_axes: Whether to decorate axes with the plotter's specific
                axis and tick labels.
            kwargs: Additional arguments to pass to :class:`BinnedPlotter`.
        """
        super().__init__(ax, bin_size=bin_size, **kwargs)

        self.block_width = int(block_width)
        self.block_size = int(block_size)

        self.coordinate_converter = BlockCoordinateConverter(
            self.block_width, self.block_size, self.bin_size, self.offset
        )

        self._style_axes()
        if decorate_axes:
            self._decorate_axes()

    def _draw_scale_bar(self) -> None:
        scale_bar_end = self.block_width / self.bin_size
        scale_bar_center = scale_bar_end / 2
        self.ax.plot(
            [0, scale_bar_end],
            [1, 1],
            ls="-",
            color="k",
            lw=2,
            transform=pad_along_y(self.ax, 0.15),
            clip_on=False,
        )
        self.ax.text(
            x=scale_bar_center,
            y=1,
            s=f"{self.block_width:,d} ISBNs",
            ha="center",
            va="bottom",
            transform=pad_along_y(self.ax, 0.2),
        )

    def _decorate_axes(
        self, xlabel: str = "Relative ISBN", ylabel: str = "Relative ISBN"
    ):
        self.ax.set_xticklabels([])

        self.ax.yaxis.set_major_formatter(
            mpl.ticker.FuncFormatter(lambda y, _: f"{y * self.block_width:.2g}")
        )

        self._draw_scale_bar()

        self.ax.set_xlabel(xlabel)
        self.ax.set_ylabel(ylabel)

        # Create a secondary axis for ISBN prefixes
        ax2 = self.ax.secondary_xaxis(
            location=1,
            functions=(
                lambda x: self.coordinate_converter.xy_to_isbn(x, 0),
                lambda x: self.coordinate_converter.isbn_to_xy(x)[0],
            ),
            transform=pad_along_y(self.ax, 0.4),
        )
        ax2.set_xlabel("ISBN prefix")
        ax2.spines[:].set_visible(False)

        ax2.xaxis.set_major_locator(mpl.ticker.MultipleLocator(self.block_size * 2))
        ax2.xaxis.set_major_formatter(
            mpl.ticker.FuncFormatter(lambda x, _: f"{str(x)[:3]}-{str(x)[3:4]}")
        )

    def define_extent(self, end_isbn: ISBN12 = LAST_ISBN) -> None:
        """Defines the extent without plotting bins or an image.

        Arguments:
            end_isbn: An end ISBN to expand an extent to from the plotter offset.
        """
        x, _ = self.coordinate_converter.isbn_to_xy(end_isbn)

        bins_per_row = math.ceil(self.block_width / self.bin_size)
        block_index = x // bins_per_row
        image_height = (self.block_size // self.bin_size) // bins_per_row
        image_width = (block_index + 1) * bins_per_row

        self._extent = (0, image_width, image_height, 0)

        x1, x2, y1, y2 = self._extent
        self.ax.set_xlim(x1, x2)
        self.ax.set_ylim(y1, y2)

        self.ax.set_aspect(self.aspect)

    def plot_bins(
        self,
        binned: BinnedArray,
        colormap: mpl.colors.Colormap | None = None,
        value_to_nan: int | None = 0,
        show_colorbar: bool = True,
        colorbar_kwargs: dict | None = None,
        **imshow_kwargs,
    ):
        self._check_bin_size(binned)
        image = rearrange_to_blocks(
            binned, binned.bin_size, self.block_width, self.block_size
        )
        return self.plot_image(
            image,
            colormap=colormap,
            value_to_nan=value_to_nan,
            show_colorbar=show_colorbar,
            colorbar_kwargs=colorbar_kwargs,
            **imshow_kwargs,
        )



"""Classes and functions to work with datasets of packed ISBN codes."""

from __future__ import annotations

import math

from dataclasses import InitVar, dataclass, field
from pathlib import Path
from struct import unpack
from typing import IO, TYPE_CHECKING, BinaryIO, NamedTuple, Self

import numpy as np
import numpy.typing as npt

from bencodepy import bread  # type: ignore[import-untyped]
from zstandard import ZstdCompressor, ZstdDecompressor

from allisbns.isbn import FIRST_ISBN, ISBN12


if TYPE_CHECKING:
    from collections.abc import Generator, KeysView


#: Packed ISBN codes that represent ISBN availability.
type PackedCodes = npt.NDArray[np.int32]

#: Unpacked representation of codes.
type UnpackedCodes = npt.NDArray[np.bool]


def load_bencoded(source: IO[bytes]) -> dict[bytes, bytes]:
    """Opens a compressed source and reads a bencoded data."""
    with ZstdDecompressor().stream_reader(source) as decompressor:
        return bread(decompressor)


def unpack_data(data: bytes) -> PackedCodes:
    """Unpacks data that come from a bencoded source."""
    return np.array(unpack(f"{len(data) // 4}I", data), dtype=np.int32)


@dataclass(frozen=True)
class BinnedArray:
    """Represents the binned data."""

    #: Bins.
    bins: npt.ArrayLike
    #: A size of bins.
    bin_size: int

    def __post_init__(self):
        object.__setattr__(self, "bins", np.asarray(self.bins))

    def __array__(
        self, dtype: npt.DTypeLike | None = None, copy: bool | None = None
    ) -> npt.NDArray:
        """Returns the underlying bins for NumPy when requested."""
        return np.asarray(self.bins, dtype=dtype, copy=copy)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.bins!r}, bin_size={self.bin_size})"


class QueryResult(NamedTuple):
    """Represents the result of a query for an ISBN.

    It shows whether the ISBN is filled (falls in a streak segment) or absent
    (in a gap segment) and the corresponding position in a segment.
    """

    #: Is the number in a streak or gap?
    is_streak: bool
    #: The index of the corresponding segment.
    segment_index: int
    #: The position in the corresponding segment.
    position_in_segment: int


@dataclass(frozen=True)
class CodeDataset:
    """Represents a dataset of the packed ISBN codes.

    Examples:
        Create a dataset from the input codes:

          >>> dataset = CodeDataset(codes=[1, 2, 3])
          >>> dataset
          CodeDataset(array([1, 2, 3], dtype=int32), bounds=(978000000000,
          978000000005))
          >>> dataset.codes
          array([1, 2, 3], dtype=int32)

        With the custom offset and fill to some ISBN:

          >>> CodeDataset(
          ...     codes=[1, 2, 3],
          ...     offset=979_000_000_000,
          ...     fill_to_isbn=979_999_999_999
          ... )
          CodeDataset(array([ 1, 2, 3, 999999994]), bounds=(979000000000,
          979999999999))
    """

    #: Packed ISBN codes.
    codes: npt.NDArray[np.int32]

    #: First ISBN in the dataset.
    offset: ISBN12 = FIRST_ISBN

    #: ISBN up to which to fill dataset codes.
    fill_to_isbn: InitVar[ISBN12 | None] = None

    #: First and last ISBNs in the dataset.
    bounds: tuple[ISBN12, ISBN12] = field(init=False)

    #: Cumulative sums of ISBNs derived from the codes.
    _isbn_cumsums: npt.NDArray = field(init=False)

    def __post_init__(self, fill_to_isbn: ISBN12 | None) -> None:
        object.__setattr__(self, "codes", np.asarray(self.codes, dtype=np.int32))

        if fill_to_isbn:
            self._fill_to_isbn(fill_to_isbn)

        object.__setattr__(
            self, "_isbn_cumsums", int(self.offset) + np.cumsum(self.codes)
        )
        object.__setattr__(
            self, "bounds", (self.offset, int(self._isbn_cumsums[-1] - 1))
        )

    def _fill_to_isbn(self, isbn: ISBN12) -> None:
        if isbn < self.offset:
            raise ValueError("isbn is less than offset")

        code_sum = np.sum(self.codes)
        if isbn < code_sum:
            raise ValueError("isbn is less than the dataset right bound")
        if isbn == code_sum:
            return

        # Check if codes end with a gap segment
        deficiency = int(isbn - self.offset - code_sum + 1)
        if len(self.codes) % 2 == 0:
            self.codes[-1] += deficiency
        else:
            object.__setattr__(
                self, "codes", np.concatenate([self.codes, [deficiency]])
            )

    @classmethod
    def from_file(
        cls,
        source: str | Path | IO[bytes],
        collection: str,
        offset: ISBN12 = FIRST_ISBN,
        fill_to_isbn: ISBN12 | None = None,
    ) -> Self:
        """Creates a dataset from a source file or byte stream.

        Arguments:
            source: A path to a bencoded compressed file or a byte stream.
            collection: A collection name to be read (e.g., 'md5', 'rgb',
                etc.). Refers to ``aarecord_id_prefix`` in the original data
                format.
            offset: The first ISBN.
            fill_to_isbn: An ISBN up to which to fill dataset codes.

        Returns:
            A dataset created from a source.
        """
        if isinstance(source, (str, Path)):
            with open(source, "rb") as f:
                input_data = load_bencoded(f)
        else:
            input_data = load_bencoded(source)
        try:
            codes = unpack_data(input_data[collection.encode()])
        except KeyError as exc:
            message = (
                f"'{collection}' not present. "
                f"Available collections: {list(input_data.keys())}"
            )
            raise ValueError(message) from exc
        return cls(codes=codes, offset=offset, fill_to_isbn=fill_to_isbn)

    @classmethod
    def from_unpacked(
        cls, unpacked_codes: UnpackedCodes, offset: ISBN12 = FIRST_ISBN
    ) -> Self:
        """Creates a dataset from the unpacked codes.

        Arguments:
            unpacked_codes: An array of unpack codes (boolean values).
            collection: A dataset collection name.
            offset: The first ISBN.

        Returns:
            A dataset created from the unpacked codes.
        """
        # This marks elements where the values change, i.e. the starts of the segments.
        change_mask = np.zeros(len(unpacked_codes) + 1, dtype=np.bool)
        change_mask[[0, -1]] = True
        np.not_equal(unpacked_codes[:-1], unpacked_codes[1:], out=change_mask[1:-1])
        packed_codes = np.diff(
            np.flatnonzero(change_mask),
            # If data starts with a gap, then prepend the packed codes with 0.
            prepend=np._NoValue if unpacked_codes[0] else [unpacked_codes[0]],
        )
        return cls(codes=packed_codes, offset=offset)

    def unpack_codes(self) -> UnpackedCodes:
        """Unpacks codes into boolean values.

        Returns:
            An array of unpacked codes.
        """
        tiles = np.zeros(len(self.codes), dtype=bool)
        tiles[::2] = True
        return np.repeat(tiles, self.codes)

    def crop(self, start: ISBN12, end: ISBN12) -> Self:
        """Crops the dataset to the input start and end ISBNs.

        Arguments:
            start: An ISBN to crop the dataset from.
            end: An ISBN to crop the dataset until.

        Returns:
            A new cropped dataset.
        """
        if start > end:
            raise ValueError("start is ahead of end")

        try:
            is_start_streak, start_idx, _ = self.query_isbn(start)
        except ValueError:
            start = self.bounds[0]
            is_start_streak = self.codes[0] != 0
            start_idx = 0
        try:
            _, end_idx, end_position = self.query_isbn(end)
        except ValueError:
            end = self.bounds[1]
            end_idx = len(self.codes) - 1
            end_position = self.codes[end_idx] - 1

        if start_idx == end_idx:
            segment_length = end - start + 1
            cropped = np.array([segment_length], self.codes.dtype)
        else:
            cropped = self.codes[start_idx : end_idx + 1].copy()
            cropped[0] = self._isbn_cumsums[start_idx] - start
            cropped[-1] = end_position + 1

        if not is_start_streak:
            cropped = np.concatenate([[0], cropped])

        return self.__class__(cropped, offset=start)

    def query_isbn(self, isbn: ISBN12) -> QueryResult:
        """Queries if the ISBN is filled in the dataset and its position.

        Arguments:
            isbn: An ISBN to query for.

        Returns:
            A query result.
        """
        if not (self.bounds[0] <= isbn <= self.bounds[1]):
            raise ValueError("input ISBN is outside of the data range")

        segment_index = int(np.searchsorted(self._isbn_cumsums, isbn, side="right"))
        is_streak = bool(segment_index % 2 == 0)
        position_in_segment = int(
            isbn - self._isbn_cumsums[segment_index] + self.codes[segment_index]
        )

        return QueryResult(is_streak, segment_index, position_in_segment)

    def check_isbns(self, isbns: list[ISBN12]) -> npt.NDArray[np.bool]:
        """Checks if ISBNs are filled in the dataset or not.

        Arguments:
            isbns: A list with ISBNs to check.

        Returns:
            An array of boolean values.
        """
        isbns_array = np.array(isbns)
        # Check if all input ISBNs are inside data range
        outside_mask = (isbns_array < self.bounds[0]) | (isbns_array > self.bounds[1])
        if np.any(outside_mask):
            raise ValueError(
                f"some ISBNs are outside of data range: {isbns_array[outside_mask]}"
            )

        segment_indices = np.searchsorted(self._isbn_cumsums, isbns_array, side="right")

        return segment_indices % 2 == 0

    def get_filled_isbns(self) -> npt.NDArray:
        """Gets filled ISBNs in the dataset.

        Returns:
            An array of filled ISBNs.
        """
        return np.flatnonzero(self.unpack_codes()) + self.offset

    def count_filled_isbns(self) -> int:
        """Counts the number of filled ISBNs in the dataset."""
        return int(np.sum(self.codes[::2]))

    def bin(
        self,
        bin_size: int = 2500,
        num_chunks: int = 4,
    ) -> BinnedArray:
        """Performs a fixed-size binning of codes into bins.

        The bin value is the number of filled ISBN values.

        Arguments:
            bin_size: A number of ISBNs in one bin.
            num_chunks: A number of chunks used to process the dataset.

        Returns:
            A binned array.
        """
        num_bins = math.ceil((self.bounds[1] - self.bounds[0] + 1) / bin_size)
        num_bins_per_chunk = math.ceil(num_bins / num_chunks)
        num_isbns_per_chunk = num_bins_per_chunk * bin_size

        all_bins = np.zeros(num_bins_per_chunk * num_chunks, dtype=np.int32)

        for chunk_idx in range(num_chunks):
            chunk_start_isbn = chunk_idx * num_isbns_per_chunk + self.offset
            chunk_end_isbn = chunk_start_isbn + num_isbns_per_chunk - 1
            unpacked_codes = self.crop(chunk_start_isbn, chunk_end_isbn).unpack_codes()
            unpacked_length = len(unpacked_codes)
            if unpacked_length != num_isbns_per_chunk:
                unpacked_codes = np.pad(
                    unpacked_codes,
                    (0, num_isbns_per_chunk - unpacked_length),
                    constant_values=0,
                )
            chunk_shift = chunk_idx * num_bins_per_chunk
            all_bins[chunk_shift : chunk_shift + num_bins_per_chunk] = np.sum(
                unpacked_codes.reshape((num_bins_per_chunk, bin_size)), axis=1
            )

        return BinnedArray(all_bins[:num_bins], bin_size)

    def is_subset(self, other: Self) -> bool:
        """Determines if a dataset is a subset of another."""
        if self.bounds[0] < other.bounds[0] or self.bounds[1] > other.bounds[1]:
            return False
        return self == other.crop(*self.bounds)

    def write_bencoded(
        self, file: str | Path | IO[bytes], prefix: str, normalize: bool = True
    ) -> None:
        """Writes ISBN codes to a bencoded compressed file.

        Arguments:
            file: A file path or file-like object to which the codes will be
                written.
            prefix: A dataset collection name. Refers to 'aarecord_id_prefix'
                in the original data format description.
            normalize: Whether to normalize codes so that the ending gap code,
                if present, is omitted (default) or not.
        """
        output_stream: IO[bytes] | BinaryIO
        if isinstance(file, (str | Path)):
            output_stream = open(file, "wb")  # noqa: SIM115
            needs_closing = True
        else:
            output_stream = file
            needs_closing = False

        try:
            codes_to_write = self.codes.copy()
            if normalize and len(self.codes) % 2 == 0:
                codes_to_write = codes_to_write[:-1]

            # Convert the codes as little-endian 4-bytes unsigned integers to bytes
            codes_as_bytes = codes_to_write.astype("<u4").tobytes()

            compressor = ZstdCompressor(level=22, threads=-1)
            with compressor.stream_writer(output_stream, closefd=False) as writer:
                writer.write(b"d")
                writer.write(f"{len(prefix)}:{prefix}".encode())
                writer.write(f"{len(codes_as_bytes)}:".encode())
                writer.write(codes_as_bytes)
                writer.write(b"e")
        finally:
            if needs_closing:
                output_stream.close()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return np.array_equal(self.codes, other.codes) and self.offset == other.offset

    def __hash__(self) -> int:
        return hash((self.codes, self.offset))

    def __getitem__(self, key: slice[ISBN12, ISBN12, None]) -> Self:
        if isinstance(key, slice):
            if key.step:
                raise ValueError("slice step is not supported")
            return self.crop(key.start, key.stop)
        raise ValueError("Object supports only slicing")

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.codes!r}, "
            f"bounds=({self.bounds[0]}, {self.bounds[1]}))"
        )


def iterate_datasets(
    data: dict[bytes, bytes],
    collections: list[str] | None = None,
    fill_to_isbn: ISBN12 | None = None,
) -> Generator[CodeDataset]:
    """Iterates over datasets created from the loaded bencoded data.

    By default, iterates over all collections in the data.

    Arguments:
        data: Loaded bencoded data that contain collections to unpack.
        collections: Collection names to unpack (e.g. 'md5', 'rgb', ...). When
            `None` (default), iterates over all collections in the data.
        fill_to_isbn: An ISBN up to which to fill dataset codes.

    Returns:
        Yields datasets for the selected collections.

    Example:
        Iterate over all datasets and count filled ISBNs::

          from allisbns.dataset import load_bencoded, iterate_datasets

          with open("aa_isbn13_codes_20251118T170842Z.benc.zst", "rb") as f:
              input_data = load_bencoded(f)

          filled_counts: dict[str, int] = {}

          collections = [x.decode() for x in input_data.keys()]
          dataset_iterator = iterate_datasets(input_data, collections)
          for collection, dataset in zip(collections, dataset_iterator):
              filled_counts[collection] = dataset.count_filled_isbns()
    """
    keys: list[bytes] | KeysView[bytes]
    if collections:
        keys = [x.encode() for x in collections]
    else:
        keys = data.keys()

    for key in keys:
        yield CodeDataset(
            codes=unpack_data(data[key]),
            fill_to_isbn=fill_to_isbn,
        )



"""Rearranges bins to binned images in various ways."""

import math

import numpy as np

from numpy.typing import ArrayLike, NDArray


def rearrange_to_rows(
    bins: ArrayLike, bin_size: int, width: int, pad_value: int = 0
) -> NDArray:
    """Rearranges bins into rows with a fixed width.

    Arguments:
        bins: An array of bins.
        bin_size: A size of bins.
        width: A width of rows.
        pad_value: A value used to fill the remaining space of the final row
            (when the number of bins is not divisible by width).

    Returns:
        Bins rearranged to rows with shape (*height*, *row width*).

    Example:
        Here is a fictitious example to demonstrate the filling order:

          >>> from allisbns.dataset import BinnedArray
          >>> binned = BinnedArray(range(8), bin_size=1)
          >>> rearrange_to_rows(
          ...      binned, bin_size=binned.bin_size, width=5
          ... )
          array([[0, 1, 2, 3, 4],
                 [5, 6, 7, 0, 0]])
    """
    if width % bin_size != 0:
        raise ValueError(
            f"row width ({width}) is not divisible by bin size ({bin_size})"
        )

    bins = np.asarray(bins)

    bins_in_row = width // bin_size
    height = math.ceil(len(bins) / bins_in_row)

    if (pad_width := bins_in_row * height - len(bins)) > 0:
        padded_bins = np.concatenate([bins, [pad_value] * pad_width])
    else:
        padded_bins = bins

    new_shape: tuple[int, int] | tuple[int, int, int]
    if bins.ndim == 1:
        new_shape = (height, bins_in_row)
    elif bins.ndim == 2:
        new_shape = (height, bins_in_row, bins.shape[1])

    return padded_bins.reshape(new_shape)


def rearrange_to_blocks(
    bins: ArrayLike,
    bin_size: int,
    block_width: int = int(1e5),
    block_size: int = int(5e7),
) -> NDArray:
    """Rearranges bins into blocks of fixed width and height.

    The arrangement is inspired by the "bookshelf" space-filling curve [1], but
    we treat it in a simpler way. So, it is basically an equivalent to the
    line-filling curve as in :meth:`rearranges_to_rows` with additional division
    into vertical blocks of the fixed size stacked horizontally afterwards.

    Arguments:
        bins: An array of bins.
        bin_size: A size of bins.
        block_width: A width of one block.
        block_size: A number of ISBNs in one block.
        pad_value: A value used to fill the remaining space of the final block
            (when the number of bins is not divisible by the block size).

    Returns:
        Bins rearranged into shape (*block height*, *block width * block count*).

    References:
        1. https://phiresky.github.io/blog/2025/visualizing-all-books-in-isbn-space/

    Example:
        Here is a fictitious example to demonstrate the filling order:

          >>> from allisbns.dataset import BinnedArray
          >>> binned = BinnedArray(range(18), bin_size=1)
          >>> rearrange_to_blocks(
          ...     binned, binned.bin_size, block_width=3, block_size=9
          ... )
          BinnedArray([[ 0,  1,  2,  9, 10, 11],
                       [ 3,  4,  5, 12, 13, 14],
                       [ 6,  7,  8, 15, 16, 17]])
    """
    if block_width % bin_size != 0:
        raise ValueError(
            f"block width ({block_width}) is not divisible by bin size ({bin_size})"
        )

    bins = np.asarray(bins)

    bins_in_row = (block_width + bin_size - 1) // bin_size
    rows_in_block = (block_size + block_width - 1) // block_width
    blocks = bins.reshape((-1, rows_in_block, bins_in_row))

    return blocks.transpose(1, 0, 2).reshape(rows_in_block, -1)



"""Classes and functions to work with ISBNs."""

from __future__ import annotations

import re

from dataclasses import asdict, dataclass
from re import Pattern
from typing import TYPE_CHECKING, Final, Literal, Self, cast

import numpy as np

from allisbns.errors import (
    InvalidISBNError,
    InvalidISBNReason,
    NotISBN12Error,
    NotISBNError,
    UndefinedISBNRangeError,
)
from allisbns.ranges import REGISTRANT_RANGES, REGISTRATION_GROUPS


if TYPE_CHECKING:
    import sys

    if "sphinx.ext.autodoc" in sys.modules:
        from collections.abc import Generator

# Type aliases

#: A valid ISBN-12 integer number.
type ISBN12 = int | np.integer

#: A string that is expected to be an ISBN.
type MaybeISBN = str

BooklandElement = Literal["978", "979"]

# Constants

#: A simple pattern to match ISBN-10 and ISBN-13 values.
ISBN_PATTERN: Pattern[str] = re.compile(r"^(?:978|979)?\d{9}[\dXx]$")

#: First defined ISBN-12.
FIRST_ISBN: Final[ISBN12] = 978_000_000_000

#: Last defined ISBN-12.
LAST_ISBN: Final[ISBN12] = 979_999_999_999

#: Total number of ISBNs.
TOTAL_ISBNS: Final = LAST_ISBN - FIRST_ISBN + 1

#: The length of ISBN-12 numbers.
ISBN12_LENGTH = 12


class CanonicalISBN:
    """Represents a canonical (normalized) ISBN-13 string.

    Only the ISBN-10 and ISBN-13 :data:`pattern <ISBN_PATTERN>` check is
    performed upon creation. So canonical ISBN values are assumed to be valid
    even if: the registration group or registrant range are undefined, the
    correct digit is incorrectly calculated. For validation, see
    :func:`validate_isbn`.

    See Also:
        :func:`normalize_isbn`
    """

    def __init__(self, value: MaybeISBN):
        """Creates a canonical ISBN-13 object."""
        self._value = self._normalize_value(value)

    @classmethod
    def _from_normalized(cls, isbn: str) -> Self:
        """Creates a new object from an already normalized ISBN-13 string."""
        new = cls.__new__(cls)
        new._value = isbn
        return new

    def _normalize_value(self, value: MaybeISBN) -> str:
        # Remove dashes and whitespaces
        output_value = _clean_isbn_string(value)

        # Check if value is ISBN-10 or ISBN-13
        if not match_isbn(output_value):
            raise NotISBNError(value)

        # Convert ISBN-10 to ISBN-13 if needed
        if len(output_value) == 10:
            output_value = f"978{output_value}"

        # Ensure that the 'X' check digit is uppercase
        if output_value[-1] == "x":
            output_value = output_value.upper()

        return output_value

    @property
    def bookland(self) -> BooklandElement:
        """Returns the bookland element."""
        return self._value[:3]  # type: ignore[return-value]

    @property
    def check_digit(self) -> str:
        """Returns the check digit."""
        return self._value[-1]

    def complete(self) -> Self:
        correct_check_digit = calculate_check_digit(self)
        if self.check_digit == correct_check_digit:
            return self
        return self._from_normalized(f"{self._value[:-1]}{correct_check_digit}")

    def hyphenate(self) -> str:
        return MaskedISBN.from_canonical(self).hyphenate()

    def to_isbn12(self) -> ISBN12:
        """Converts to an :data:`ISBN12` number."""
        return int(self._value[:ISBN12_LENGTH])

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self._value == other._value

    def __hash__(self) -> int:
        return hash(self._value)

    def __getitem__(self, key: int | slice) -> str:
        return self._value[key]

    def __str__(self) -> str:
        return self._value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._value})"


@dataclass(frozen=True)
class MaskedISBN:
    """Represents a masked structure of an ISBN-13 string.

    Example:
        It is more practical to create it from a canonical ISBN:
          >>> canonical = normalize_isbn("978-000-000-000-X")
          >>> MaskedISBN.from_canonical(canonical)
          MaskedISBN(
              bookland='978',
              group='0',
              registrant='00',
              publication='000000',
              check_digit='X',
          )
    """

    #: Bookland, or GS1 element.
    bookland: BooklandElement
    #: Registration group element.
    group: str
    #: Registrant element.
    registrant: str
    #: Publication element.
    publication: str
    #: Check digit.
    check_digit: str

    @classmethod
    def from_canonical(cls, isbn: CanonicalISBN) -> Self:
        """Creates a masked structure from a canonical ISBN-13 string."""
        max_registrant_length = 7

        bookland = isbn.bookland
        current_index = len(bookland)

        # Extract a registration group element
        group_prefix = extract_group_prefix(isbn)
        _, group = group_prefix.split("-")
        current_index += len(group)

        # Match a registrant element
        isbn_rest_part = isbn[current_index:]
        registrant_ranges = REGISTRANT_RANGES[group_prefix]
        for range_start, range_end, registrant_length in registrant_ranges:
            registrant_probe = isbn_rest_part[:max_registrant_length]
            # For some groups (e.g., '978-99998'), the length of ISBN-13 may not
            # be enough to get the probe
            registrant_probe = registrant_probe.ljust(max_registrant_length, "0")
            if range_start <= registrant_probe <= range_end:
                if registrant_length == 0:
                    message = (
                        "not defined for use registrant: "
                        f"'{group_prefix}-{registrant_probe}'"
                    )
                    raise UndefinedISBNRangeError(
                        message, InvalidISBNReason.BAD_REGISTRANT, isbn
                    )
                registrant = isbn_rest_part[:registrant_length]
                current_index += registrant_length
                break

        return cls(
            bookland=bookland,
            group=group,
            registrant=registrant,
            publication=isbn[current_index:-1],
            check_digit=isbn.check_digit,
        )

    @property
    def elements(self) -> dict[str, str | BooklandElement]:
        return asdict(self)

    def hyphenate(self) -> str:
        """Formats an ISBN string with elements delimited by a hyphen."""
        return "-".join(self.elements.values())

    def __getitem__(self, key: int | slice[int, int, None]) -> str:
        match key:
            case slice() as x if x.step is not None:
                raise ValueError("slice step is not supported")
            case slice() as x:
                return "-".join(list(self.elements.values())[x])
            case int() as index:
                return list(self.elements.values())[index]
            case _:
                raise TypeError("index must be integer or slice object")


def _clean_isbn_string(x: MaybeISBN) -> MaybeISBN:
    return x.replace("-", "").replace(" ", "")


def ensure_isbn12(value: int | np.integer) -> ISBN12:
    """Ensures that a value is an ISBN-12 number."""
    if not isinstance(value, (int, np.integer)):
        raise TypeError(f"value must be integer, got {type(value)}")
    if not (FIRST_ISBN <= value <= LAST_ISBN):
        raise NotISBN12Error(
            f"value must be in range [{FIRST_ISBN}, {LAST_ISBN}], got {value}"
        )
    return value


def calculate_check_digit(isbn: CanonicalISBN | ISBN12) -> int:
    """Calculates the check digit for an ISBN."""
    isbn_value = str(isbn)[:ISBN12_LENGTH]
    total = 0
    for i in range(len(isbn_value)):
        digit = int(isbn_value[i])
        if i % 2 == 0:
            total += digit
        else:
            total += digit * 3
    return (10 - (total % 10)) % 10


def normalize_isbn(value: MaybeISBN | ISBN12, correct: bool = False) -> CanonicalISBN:
    """Normalizes an ISBN to a canonical form.

    Arguments:
        value: An value to normalize.
        correct: Whether to check and replace the incorrect check digit with the
            correct one. Defaults to `False`.

    Examples:
        1) The input ISBN-13 string with the correct check digit:

          >>> normalize_isbn("978-000-000-000-2")
          CanonicalISBN(9780000000002)

        2) The input ISBN-13 string with the incorrect check digit:

          >>> # Keep it as it is
          >>> normalize_isbn("978-000-000-000-3")
          CanonicalISBN(9780000000003)

          >>> # Or, correct the check digit
          >>> normalize_isbn("978-000-000-000-3", correct=True)
          CanonicalISBN(9780000000002)

        3) The input :data:`ISBN12` integer value:

          >>> value: ISBN12 = 978_000_000_000
          >>> normalize_isbn(value)
          CanonicalISBN(9780000000002)
    """
    match value:
        case int() | np.integer():
            isbn12 = ensure_isbn12(value)
            normalized_isbn = CanonicalISBN._from_normalized(f"{isbn12}X").complete()
        case str():
            normalized_isbn = CanonicalISBN(cast("str", value))
            if normalized_isbn.check_digit != "X" and correct:
                normalized_isbn = normalized_isbn.complete()
        case _:
            pass
    return normalized_isbn


def match_isbn(x: MaybeISBN) -> bool:
    """Checks that the input string is an ISBN-10 or ISBN-13."""
    return bool(re.match(ISBN_PATTERN, _clean_isbn_string(x)))


def validate_isbn(
    value: CanonicalISBN, *, return_reasons: bool = False
) -> bool | tuple[bool, list[InvalidISBNReason]]:
    """Validates a canonical ISBN.

    The validation includes checks for: (1) an undefined registration group, (2) an
    undefined registrant, and (3) an incorrect check digit. To check if a string is an
    ISBN or not, see :meth:`match_isbn` instead.

    Arguments:
        value: A canonical ISBN to validate.
        return_reasons: Whether to return reasons for failed checks or not.

    Returns:
        A boolean value or, if ``return_reasons`` is requested, a boolean value
        and a list of reasons for failed checks.

    See Also:
        :meth:`match_isbn`

    Examples:
        If you need to check a list of unsure ISBNs for validity, you can
        combine this function with :meth:`match_isbn`::

          def check_isbn_for_errors(x: MaybeISBN) -> list[InvalidISBNReason]:
              if not match_isbn(x):
                  return [InvalidISBNReason.NOT_ISBN]
              else:
                  canonical_isbn = normalize_isbn(x)
                  _, errors = validate_isbn(canonical_isbn, return_reasons=True)
                  return errors
    """
    if not isinstance(value, CanonicalISBN):
        raise TypeError("value should be of type 'CanonicalISBN'")

    errors: list[InvalidISBNReason] = []

    # Check if the ISBN contains a correct check digit
    current_check_digit = value[-1]
    if current_check_digit.isnumeric():
        calculated_check_digit = calculate_check_digit(value)
        if current_check_digit != str(calculated_check_digit):
            if return_reasons:
                errors.append(InvalidISBNReason.BAD_CHECK_DIGIT)
            else:
                return False

    # Check if the ISBN is of a valid registration group and not in any
    # undefined registrant ranges
    try:
        MaskedISBN.from_canonical(value)
    except InvalidISBNError as e:
        if return_reasons:
            errors.append(e.reason)
        else:
            return False

    if return_reasons:
        return not bool(errors), errors

    return True


def get_prefix_bounds(prefix: str) -> tuple[ISBN12, ISBN12]:
    """Gets the bounds for an ISBN prefix."""
    prefix = prefix.replace("-", "")
    return (
        int(prefix.ljust(ISBN12_LENGTH, "0")),
        int(prefix.ljust(ISBN12_LENGTH, "9")),
    )


def get_prefix_capacity(prefix: str) -> int:
    """Gets the capacity of an ISBN prefix."""
    return 10 ** (ISBN12_LENGTH - len(prefix.replace("-", "")))


def extract_group_prefix(isbn: CanonicalISBN) -> str:
    """Extracts the group prefix from the ISBN.

    Arguments:
        isbn: A canonical ISBN-13 string.

    Returns:
        A found full ISBN group prefix, e.g. '978-1'.

    Raises:
        :class:`~allisbns.errors.InvalidISBNError`: If the ISBN does not
        contain valid group prefix.

    Note:
        Group ranges were periodically updated from the RangeMessage.xml file
        available from the International ISBN Agency and may not be the most
        recent.

    See Also:
        :class:`MaskedISBN`
    """
    bookland_length = 3
    max_group_length = 5

    isbn_string = str(isbn)

    for group_length in range(1, max_group_length + 1):
        group = isbn_string[bookland_length : bookland_length + group_length]
        group_prefix = f"{isbn_string[:bookland_length]}-{group}"
        if group_prefix in REGISTRATION_GROUPS:
            break
    else:
        raise UndefinedISBNRangeError(
            f"invalid (or unknown) registration group: {group_prefix}",
            InvalidISBNReason.BAD_GROUP,
            isbn,
        )

    return group_prefix


def generate_isbns(
    start_isbn: ISBN12 | None = None, limit: int | None = None
) -> Generator[CanonicalISBN]:
    """Yields canonical ISBN-13 string values.

    Arguments:
        start_isbn: A start ISBN-12 number. Defaults to `None`, meaning yields
            from the first ISBN-13, '978-000-000-000-2'.
        limit: A maximum number of ISBNs to yields. Defaults to `None`, meaning
            yields till the last ISBN-13, '979-999-999-999-0'.

    """
    if start_isbn is None:
        start_isbn = FIRST_ISBN
    start_isbn = int(start_isbn)

    if not (FIRST_ISBN <= start_isbn <= LAST_ISBN):
        raise ValueError(
            f"start value must be in range [{FIRST_ISBN:_d}, {LAST_ISBN:_d}], "
            f"got {start_isbn:_d}"
        )

    if limit is not None:
        end_isbn: ISBN12 = start_isbn + limit - 1
        end_isbn = min(end_isbn, LAST_ISBN)
    else:
        end_isbn = LAST_ISBN

    for isbn in range(int(start_isbn), int(end_isbn) + 1):
        yield normalize_isbn(isbn)
